# -*- coding: utf-8 -*-
"""
Created on Tue May 31 10:08:29 2022

@author: Jinhai
"""

import numpy as np
import sys
from DigiNet.layers.activations.sigmoid_hdw import sigmoid
from DigiNet.layers.activations.tanh_hdw import tanh

initial = {
    16:{
        4: '0000',
        }
    }

class LSTM_hdw:
    
    def __init__(
            self, 
            input_len, 
            output_len,
            Quantization = [16, 4], 
            fract_width = 8,
            W_in = 'W_in.txt', 
            W_rec = 'W_rec.txt', 
            Bias = 'Bias.txt',
            **kwargs,
            ):
        
        try:
            self.W_in  = np.loadtxt(W_in, dtype='str')
            self.W_rec = np.loadtxt(W_rec, dtype='str')
            self.Bias  = np.loadtxt(Bias, dtype='str')
        except:
            print('Weight path is not correct!')
        
        self.input_len  = input_len
        self.output_len = output_len
        self.Quantization = Quantization
        self.fract_width = fract_width
        
        [self.W_f, self.W_g, self.W_i, self.W_o] = np.array([self.W_in]).reshape(4, self.output_len, self.input_len)
        [self.R_f, self.R_g, self.R_i, self.R_o] = np.array([self.W_rec]).reshape(4, self.output_len, self.output_len)
        [self.B_f, self.B_g, self.B_i, self.B_o] = np.array([self.Bias]).reshape(4, self.output_len)
        
        
    def Forward(self, X):
        self.X = X
        [t, dim] = np.shape(X)
        for i in range (t):
            if i == 0:
                h_ini = np.full(self.output_len, initial[16][4])
                c_ini = np.full(self.output_len, initial[16][4])
                self = LSTM_hdw.SFE(self, c_ini, h_ini, self.X[i])
            else:
                self = LSTM_hdw.SFE(self, self.c_out, self.h_out, self.X[i])
                
        return self
        
    
    # Single Forward Element
    def SFE(self, c_in, h_in, X_in):
        # time step one
        z_f = LSTM_hdw.VMM(self, X_in, self.W_f, h_in, self.R_f, self.B_f)
        z_i = LSTM_hdw.VMM(self, X_in, self.W_i, h_in, self.R_i, self.B_i)
        z_g = LSTM_hdw.VMM(self, X_in, self.W_g, h_in, self.R_g, self.B_g)
        z_o = LSTM_hdw.VMM(self, X_in, self.W_o, h_in, self.R_o, self.B_o)
        
        # time step two
        T_f = sigmoid(self, z_f).val
        T_i = sigmoid(self, z_i).val
        T_g = tanh(self, z_g).val
        T_o = sigmoid(self, z_o).val
        
        # time step three --> Hadamard Product
        P_f = LSTM_hdw.Hadamard(self, c_in, T_f)
        P_ig = LSTM_hdw.Hadamard(self, T_i, T_g)
        
        # time step four
        c_out = P_f + P_ig
        
        # time step five
        T_c = tanh(self, c_out).val
        
        # time step six
        h_out = LSTM_hdw.Hadamard(self, T_c, T_o)
        
        # save parameters
        self.z_f, self.z_i, self.z_g, self.z_o = z_f, z_i, z_g, z_o
        self.T_f, self.T_i, self.T_g, self.T_o = T_f, T_i, T_g, T_o
        self.P_f, self.P_ig = P_f, P_ig
        if self.Quantization[0] == 16:
            self.c_out = tuple([hex(m) for m in c_out])
        self.T_c = T_c
        if LSTM_hdw.Quantization[0] == 16:
            self.h_out = tuple([hex(n) for n in h_out])
        
        return self
        
    
    def VMM(self, X, W_in, h_in, W_rec, Bias):
        dim_in = np.size(X)
        dim_rec = np.size(h_in)
        if dim_in != np.shape(W_in)[1] or dim_rec != np.shape(W_rec)[1]:
            print("I/W dimension do not match")
            sys.exit()
            
        out = np.zeros(self.output_len, dtype='int32')
        p1 = out.copy()
        p2 = out.copy()
        dtype = self.Quantization[0]
        for i in range(dim_rec):
            for j in range(dim_in):
                p1[i] += int(W_in[i, j], dtype) * int(X[j], dtype) >> self.fract_width
            for k in range(dim_rec):
                p2[i] += int(W_rec[i, k], dtype) * int(h_in[k], dtype) >> self.fract_width
            out[i] = p1[i] + p2[i] + int(Bias[i], dtype)
        return out
    
    
    def Hadamard(self, arr1, arr2):
        if np.size(arr1) != np.size(arr2):
            print("Dimension do not match")
            sys.exit()
        
        out = np.zeros(self.output_len, dtype='int32')
        dim = np.size(arr1)
        dtype = self.Quantization[0]
        for i in range (dim):
            out[i] = int(arr1[i], dtype) * int(arr2[i], dtype) >> self.fract_width
        return out
    